from .plot import Plot

from bokeh.embed import file_html
from bokeh.layouts import column, row
from bokeh.models.widgets import Div
from bokeh.plotting import show as show_bokeh
from bokeh.resources import CDN
import numpy as np


def _make_plot(plot, width_total_as_session, share_x, share_y):
    # TODO: Check the shape first, the types etc
    def build_plot(plot, width=None, x_range=None, y_range=None):
        fig = plot.make_figure()
        if share_x:
            if x_range is None:
                x_range = fig.x_range
            else:
                fig.x_range = x_range
        if share_y:
            if y_range is None:
                y_range = fig.y_range
            else:
                fig.y_range = y_range
        if width:
            fig.width = width
        if not plot.grid_visible:
            fig.xgrid.visible = False
            fig.ygrid.visible = False
        for step in plot.steps:
            step(fig)
        if width:
            div = Div(text=plot.description, width=width, height=None)
        else:
            div = Div(text=plot.description, width=plot.width, height=None)
        return [column(fig, div), x_range, y_range]

    if isinstance(plot, Plot):
        if width_total_as_session:
            return build_plot(plot=plot, width=plot.width_session)[0]
        else:
            return build_plot(plot)[0]
    elif isinstance(plot, (list, np.ndarray)):
        x_range = None
        y_range = None
        row_all = []
        for p_1 in plot:
            if isinstance(p_1, Plot):
                width = p_1.width_session if width_total_as_session else None
                if x_range is None:
                    if y_range is None:
                        f_r, xr, yr = build_plot(p_1, width=width)
                        x_range = xr
                        y_range = yr
                    else:
                        f_r, xr, yr = build_plot(p_1, width=width,
                                                 y_range=y_range)
                        x_range = xr
                else:
                    if y_range is None:
                        f_r, xr, yr = build_plot(p_1, width=width,
                                                 x_range=x_range)
                        y_range = yr
                    else:
                        f_r = build_plot(p_1, width=width, x_range=x_range,
                                         y_range=y_range)[0]
                row_all.append(f_r)
            elif isinstance(p_1, (list, np.ndarray)):
                fig_width = int(p_1[0].width_session / len(p_1))
                width = fig_width if width_total_as_session else None
                row_i = []
                for pp in p_1:
                    if x_range is None:
                        if y_range is None:
                            f_r, xr, yr = build_plot(pp, width=width)
                            x_range = xr
                            y_range = yr
                        else:
                            f_r, xr, yr = build_plot(pp, width=width,
                                                     y_range=y_range)
                            x_range = xr
                    else:
                        if y_range is None:
                            f_r, xr, yr = build_plot(pp, width=width,
                                                     x_range=x_range)
                            y_range = yr
                        else:
                            f_r, xr, yr = build_plot(pp, width=width,
                                                     x_range=x_range,
                                                     y_range=y_range)
                    row_i.append(f_r)
                row_all.append(row(row_i))
        return column(row_all)


def show_base(plot, width_total_as_session=False, share_x=False,
              share_y=False):
    show_bokeh(_make_plot(plot=plot,
                          width_total_as_session=width_total_as_session,
                          share_x=share_x, share_y=share_y))


def _update_show_default_args(show_base, session):
    def show_updated(plot,
                     width_total_as_session=session.width_total_as_session,
                     share_x=False, share_y=False):
        show_base(plot=plot, width_total_as_session=width_total_as_session,
                  share_x=share_x, share_y=share_y)
    return show_updated


def _update_save_default_args(save_base, session):
    def save_updated(plot, save_path=session.save_path,
                     file_exists_mode=session.file_exists_mode,
                     width_total_as_session=session.width_total_as_session,
                     share_x=False, share_y=False):
        save_base(plot=plot, save_path=save_path,
                  file_exists_mode=file_exists_mode,
                  width_total_as_session=width_total_as_session,
                  share_x=share_x, share_y=share_y)
    return save_updated


def save_base(plot, save_path, file_exists_mode, width_total_as_session,
              share_x, share_y, get_html_only):
    if not save_path:
        # TODO: Add warning
        return None

    plot_made = _make_plot(plot, width_total_as_session=width_total_as_session,
                           share_x=share_x, share_y=share_y)
    html = file_html(plot_made, CDN)

    if not save_path.endswith('.html'):
        save_path += '.html'
    if file_exists_mode.lower() == 'append':
        with open(file=save_path, mode='a',) as f:
            f.write(html)
    elif file_exists_mode.lower() == 'overwrite':
        with open(file=save_path, mode='w',) as f:
            f.write(html)


def is_color(c):
    if isinstance(c, str):
        return True
    if isinstance(c, (list, np.ndarray, tuple)) and (
            len(c) == 3) and isinstance(c[0], (float, int)):
        return True
    return False


def from_rgb_to_hex(col):
    return '#' + hex(col[0])[2:].upper().zfill(2) + hex(col[1])[
                                                    2:].upper().zfill(2) + hex(
        col[2])[2:].upper().zfill(2)


def format_color(col):
    if isinstance(col, (list, np.ndarray, tuple)) and all(
            [isinstance(c_i, (float, int)) for c_i in col]) and (
            len(col) == 3):
        if 0 <= np.min(col) <= np.max(col) <= 1:
            rgb = (int(255 * col[0]), int(255 * col[1]), int(255 * col[2]))
        else:
            rgb = (int(col[0]), int(col[1]), int(col[2]))
        return from_rgb_to_hex(rgb)
    else:
        return col


def is_iterable(obj):
    ''' Check if an object is iterable
    Cf https://stackoverflow.com/questions/1952464/in-python-how-do-i- \
    determine-if-an-object-is-iterable

    Args:
        obj (Any object)

    Returns
        is_iterable (bool)
    '''
    try:
        iter(obj)
    except Exception:
        return False
    else:
        return True



from .plot import Plot
from .tools import show_base, save_base, is_color, format_color, is_iterable
from ..tools.color_palettes import palette_from_name_to_function

import copy
import numbers

from bokeh.models import Range1d
from bokeh.plotting import figure
from bokeh.models import ColorBar, LinearColorMapper
import numpy as np
import pandas as pd


def histogram_base(x, y, source_dataframe, tick_label, label_orientation,
                   width, height, description, title, x_label, y_label,
                   show_plot, color, colorbar_type, legend, bar_width, alpha,
                   x_axis_type, y_axis_type, x_range, y_range, grid_visible,
                   session, save_path):
    """ Scatter plot

    Args:
        x (array-like): X-axis data
        y (array-like): y-axis data
    """
    # We convert (source, x and y) into only x and y. x and y will be processed
    # normally. source will not be used any more.
    if source_dataframe is not None:
        if isinstance(source_dataframe, dict):
            # TODO: If the user passes a tuple that is not a key, add a
            # message suggesting that he may want to use a list instead
            # because a tuple is considered as a key, and not as an iterable
            # of keys
            # Same for DataFrame
            if y not in source_dataframe.keys():
                raise ValueError('The value of y is not a key of the '
                                 'dictionary provided in source_dataframe')
            if isinstance(legend, str) and (legend.lower() == 'auto'):
                if isinstance(y, str):
                    legend = str(y)
            y = source_dataframe[y]

            if x not in source_dataframe.keys():
                raise ValueError('The value of x is not a key of the '
                                 'dictionary provided in source_dataframe')
            x = source_dataframe[x]

        elif isinstance(source_dataframe, pd.DataFrame):
            if y not in source_dataframe.keys():
                raise ValueError('The value of y is not a key of the '
                                 'DataFrame provided in source_dataframe')
            if isinstance(legend, str) and (legend.lower() == 'auto'):
                if isinstance(y, str):
                    legend = str(y)
            y = source_dataframe[y]

            if x is None:
                # Be careful, the behavior is different for a dict and for a
                # Dataframe
                x = source_dataframe.index.values
            else:
                if x not in source_dataframe.keys():
                    raise ValueError('The value of x is not a key of the '
                                     'DataFrame provided in source_dataframe')
                x = source_dataframe[x].values

    if y is None:
        raise ValueError('y must be specified. It must be a one dimensional '
                         'array like')
    if x is None:
        x = np.arange(len(y))
    if (not isinstance(x, (list, np.ndarray, tuple))) \
            or (not isinstance(y, (list, np.ndarray, tuple))):
        raise ValueError('X and y must be a one dimensional array like')
    if (np.ndim(x) != 1) or (np.ndim(y) != 1):
        raise ValueError('X and y must be a one dimensional array like')

    # We pre-process `color`
    _color_bar_made = False
    if color is None:
        nb_color_needed = 1
        color = palette_from_name_to_function[session.palette_name](
            nb_color_needed)
        color = [color[0] for _ in y]
    # We pre-process `color`
    # Corner case: if color = [0.5, 0.6, 0.7] and len(y) == 3, we cannot say if
    # color means actually the color defined by [0.5, 0.6, 0.7] or if this
    # quantity should be converted into categorical colors (3 colors here cause
    # the 3 values are different). For that case we use the argument
    # `automatic_color_mapping` in from the session.

    elif (len(color) == 3)\
            and isinstance(color[0], numbers.Real)\
            and session.automatic_color_mapping \
            and (len(y) == 3):
        color = [color, color, color]
    else:
        # General case
        if is_color(color):
            color = [color for _ in y]
        elif isinstance(color, (list, np.ndarray, tuple))\
                and is_color(color[0]):
            if len(color) == len(y):
                pass  # Good case
            elif len(color) == 2:
                # TODO: Implement a linear set of color between the 2 given
                #  colors
                raise ValueError('Number of elements in `color` invalid')
            else:
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')
        elif isinstance(color, (list, np.ndarray, tuple)) and isinstance(
                color[0], numbers.Real):
            if len(color) != len(y):
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')

            if not colorbar_type:
                colorbar_type = 'auto'
                _add_color_bar = False
            else:
                _add_color_bar = True

            if colorbar_type.lower() == 'auto':
                if ((len(np.unique(color)) / len(color)) <= 0.5) and (
                  len(np.unique(color) < 9)):
                    # Data are considered categorical
                    colorbar_type = 'categorical'
                else:
                    colorbar_type = 'continuous'
            else:
                colorbar_type = colorbar_type.lower()

            if colorbar_type == 'categorical':
                if isinstance(legend, str) and (legend.lower() == 'auto'):
                    legend = [str(c) for c in color]
                color_unique = sorted(list(np.unique(color)))
                nb_color_needed = len(color_unique)
                palette_colors = palette_from_name_to_function[
                    session.palette_name](nb_color_needed)
                color = [palette_colors[color_unique.index(c)] for c in color]
            elif colorbar_type == 'continuous':
                palette_colors = palette_from_name_to_function[
                    session.palette_name](256)
                color_mapper = LinearColorMapper(palette=palette_colors,
                                                 low=np.min(color),
                                                 high=np.max(color))
                color = np.array(color)
                col_indexes = (color - color.min()) / (
                            (color.max() - color.min()) / (256 - 1))
                color_min = np.min(color)
                color_max = np.max(color)
                color = [palette_colors[int(ci)] for ci in col_indexes]
                if _add_color_bar:
                    color_bar = ColorBar(color_mapper=color_mapper,
                                         location=(0, 0))
                    _color_bar_made = True
        elif isinstance(color, (list, np.ndarray, tuple)):
            if len(color) != len(y):
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')

            if isinstance(colorbar_type, str) and (
                    colorbar_type.lower() == 'continuous'):
                error_msg = (
                    'With such elements in `color`, the colorbar cannot be '
                    'continuous because it can be transformed into numbers. '
                    'Set colorbar to either `auto` or `categorical`'
                )
                raise ValueError(error_msg)
            if isinstance(legend, str) and (legend.lower() == 'auto'):
                legend = [str(c) for c in color]
            color_unique = sorted(list(np.unique(color)))
            nb_color_needed = len(color_unique)
            palette_colors = palette_from_name_to_function[
                session.palette_name](nb_color_needed)
            color = [palette_colors[color_unique.index(c)] for c in color]

    color = [format_color(c) for c in color]

    # We pre-process `legend`
    # add_legend = True
    if (legend is None) or (legend == 'auto'):
        legend = ['' for _ in y]
        # add_legend = False
    elif isinstance(legend, str):
        legend = [legend for _ in y]
    elif isinstance(legend, (list, np.ndarray, tuple)):
        if len(legend) != len(y):
            raise ValueError('The number of elements in `legend` is not '
                             'consistent with the data')

    # We pre-process `alpha`
    if isinstance(alpha, numbers.Real):
        alpha = [alpha for _ in y]
    elif isinstance(alpha, (list, np.ndarray, tuple)):
        if len(alpha) != len(y):
            raise ValueError(
                'The alpha argument given is non consistent with the data')

    # We pre-process `x_axis_type` and `y_axis_type`
    if x_axis_type.lower() == 'auto':
        if isinstance(x[0], numbers.Real):
            x_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(x[0], errors='raise')
            except ValueError:
                raise ValueError('`x_axis_type` is set to `auto` and x[0] '
                                 'is neither a number, nor an object parsable '
                                 'as a date. Object x[0]: {}'.format(x[0]))
            x_axis_type = 'datetime'
            x = pd.to_datetime(x)
    elif x_axis_type.lower() in ['numeric', 'numerical']:
        x_axis_type = 'linear'
    elif x_axis_type.lower() in ['date', 'datetime', 'time']:
        x_axis_type = 'datetime'
        x = pd.to_datetime(x)
    # Same with y axis
    if y_axis_type.lower() == 'auto':
        if isinstance(y[0], numbers.Real):
            y_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(y[0], errors='raise')
            except ValueError:
                raise ValueError('`y_axis_type` is set to `auto` and y[0] '
                                 'is neither a number, nor an object parsable '
                                 'as a date. Object y[0]: {}'.format(y[0]))
            y_axis_type = 'datetime'
            y = pd.to_datetime(y)
    elif y_axis_type.lower() in ['numeric', 'numerical']:
        y_axis_type = 'linear'
    elif y_axis_type.lower() in ['date', 'datetime', 'time']:
        y_axis_type = 'datetime'
        y = pd.to_datetime(y)

    # We pre-process `bar_width`
    if isinstance(bar_width, numbers.Real):
        bar_width = [bar_width for _ in y]
    elif isinstance(bar_width, (list, np.ndarray, tuple)):
        if len(bar_width) != len(y):
            raise ValueError(
                'The bar_width argument given is non consistent '
                'with the data')
    elif isinstance(bar_width, str) and (bar_width.lower() == 'auto'):
        if len(x) == 1:
            bar_width = [1 for _ in y]
        else:
            bar_width = [
                np.min(np.abs(np.diff(sorted(x)))) * 0.8 for _ in y
            ]
    else:
        try:
            bar_width_td = pd.to_timedelta(bar_width)
            bar_width = [bar_width_td for _ in y]
        except ValueError:
            raise ValueError('Error in the `bar_width` argument')

    # We pre-process `tick_label`
    def _format_x_val(x_val):
        if int(x_val) == x_val:
            return int(x_val)
        else:
            return float(x_val)
    if isinstance(tick_label, (list, np.ndarray, tuple)):
        if x_axis_type == 'linear':
            # Note that bokeh major_label_overrides dict will not accept
            # 1.0 to replace the label of 1
            major_label_overrides = {_format_x_val(x[i]): str(tl) for i, tl in
                                     enumerate(tick_label)}
        else:
            major_label_overrides = {_format_x_val(x[i]): str(tl) for i, tl in
                                     enumerate(tick_label)}
    elif tick_label is None:
        major_label_overrides = {}
    else:
        major_label_overrides = {_format_x_val(x_i): str(tick_label) for x_i in
                                 x}

    # We group x and y based on legend because in bokek, figure.scatter can
    # only set one legend label by scatter plot. So if the legend contains
    # an iterable of strings, we need to look for unique values and mask
    # all attributes
    legend_unique = np.unique(legend)
    mask_all = [l_i == np.array(legend) for l_i in legend_unique]
    x_copy = copy.deepcopy(x)
    x = [np.array(x)[mask] for mask in mask_all]
    y = [np.array(y)[mask] for mask in mask_all]
    color = [np.array(color)[mask] for mask in mask_all]
    bar_width = [np.array(bar_width)[mask] for mask in mask_all]
    alpha = [np.array(alpha)[mask] for mask in mask_all]

    steps = []
    legend_exist = False
    legend_unique = [str(lu) for lu in legend_unique]
    for (x_i, y_i, col_i, leg_i, bw_i, a_i) in zip(x, y, color, legend_unique,
                                                   bar_width, alpha):
        if leg_i:
            legend_exist = True

            def step(f, x_c=x_i, y_copy=y_i, col_c=col_i, leg_c=leg_i,
                     bw_c=bw_i, a_c=a_i):
                left = [x_c_i - (bw_c_i / 2) for x_c_i, bw_c_i in
                        zip(x_c, bw_c)]
                right = [x_c_i + (bw_c_i / 2) for x_c_i, bw_c_i in
                         zip(x_c, bw_c)]
                f.quad(bottom=0, top=y_copy, left=left, right=right,
                       color=col_c, legend_label=leg_c, alpha=a_c)
        else:
            def step(f, x_c=x_i, y_copy=y_i, col_c=col_i,
                     bw_c=bw_i, a_c=a_i):
                # Done sequencially because because in vectorial we cannot
                # substract numpy.int64 and pandas timedeltas
                left = [x_c_i - (bw_c_i / 2) for x_c_i, bw_c_i in
                        zip(x_c, bw_c)]
                right = [
                    x_c_i + (bw_c_i / 2) for x_c_i, bw_c_i in zip(x_c, bw_c)
                ]
                # left = np.array(x_c) - (np.array(bw_c) / 2)
                # right = np.array(x_c) + (np.array(bw_c) / 2)
                f.quad(bottom=0, top=y_copy, left=left, right=right,
                       color=col_c, alpha=a_c)
        steps.append(step)

    if legend_exist:
        def make_legend_interactive(f):
            f.legend.click_policy = "hide"
        steps.append(make_legend_interactive)

    def format_ticks(f, x_axis_type_c=x_axis_type, x_copy_c=x_copy,
                     major_label_overrides_c=major_label_overrides):
        try:
            f.xaxis.ticker.ticks = f.xaxis.ticker.ticks + x_copy_c
        except Exception as e:
            f.xaxis.ticker = x_copy_c
        try:
            f.xaxis.major_label_overrides.update(major_label_overrides_c)
        except Exception as e:
            f.xaxis.major_label_overrides = major_label_overrides_c

    if (x_axis_type == 'linear') \
            and isinstance(x_copy[0], numbers.Real) and major_label_overrides:
        steps.append(format_ticks)

    def _make_fig():
        fig = figure(width=width, height=height, title=title,
                     background_fill_color=session.background_color,
                     x_axis_label=x_label, y_axis_label=y_label,
                     x_axis_type=x_axis_type, y_axis_type=y_axis_type)
        fig.title.align = 'center'
        fig.title.text_color = '#33331a'
        fig.xgrid.grid_line_dash = [8, 3, 2, 3]
        fig.ygrid.grid_line_dash = [8, 3, 2, 3]
        fig.toolbar.autohide = True
        if x_range is not None:
            if not is_iterable(x_range):
                raise ValueError('`x_range` argument should be an iterable')
            fig.x_range = Range1d(x_range[0], x_range[1])
        if y_range is not None:
            if not is_iterable(y_range):
                raise ValueError('`y_range` argument should be an iterable')
            fig.y_range = Range1d(y_range[0], y_range[1])
        # if (x_axis_type == 'linear') and isinstance(x_copy[0], numbers.Real):
        #     if major_label_overrides:
        #         fig.xaxis.ticker = x_copy
        #         fig.xaxis.major_label_overrides = major_label_overrides
        fig.xaxis.major_label_orientation = label_orientation
        if _color_bar_made and _add_color_bar:
            color_mapper = LinearColorMapper(palette=palette_colors,
                                             low=color_min,
                                             high=color_max)
            c_bar = ColorBar(color_mapper=color_mapper, location=(0, 0))
            fig.add_layout(c_bar, 'right')
        return fig

    plot = Plot(make_figure=_make_fig, steps=steps, description=description,
                figure=_make_fig(), width=width, grid_visible=grid_visible,
                width_session=session.width, session=session)

    if save_path:
        save_base(plot=plot, save_path=save_path,
                  file_exists_mode=session.file_exists_mode,
                  width_total_as_session=False, share_x=False, share_y=False)

    if show_plot:
        return show_base(plot) or plot
    else:
        return plot


def _update_histogram_default_args(histogram_base, session):
    def histogram_updated(y, x=None, source_dataframe=None, tick_label=None,
                          label_orientation='horizontal', width=session.width,
                          height=session.height,
                          description=session.description, title=session.title,
                          x_label=None, y_label=None,
                          show_plot=session.show_plot, color=None,
                          colorbar_type='auto', legend='auto',
                          bar_width='auto', alpha=1, x_axis_type='auto',
                          y_axis_type='auto', x_range=None, y_range=None,
                          save_path=session.save_path,
                          grid_visible=session.grid_visible):
        """ Plot a graph with points (scatter-plot)

        Args:
            y (array-like of dimension 1 or dict key or pd.dataframe col name):
                The height of each bar in the histogram. The values can be
                either numbers, or dates (datetimes or parsable strings).
                If a `source_dataframe` is given, `x` and `y` should be column
                names / keys.

            x (array-like of dim 1 or None or dict key or dataframe col name):
                The x-coordinates of the center of the bars of the histogram.
                If None and source_dataframe not is a pandas dataframe:
                `range(len(y))` is used as `x`.
                If None and source_dataframe is a pandas dataframe: the index
                of the dataframe is used as `x`.
                The values can be either numbers, or dates (datetimes or
                parsable strings). If a `source_dataframe` is given, `x` and
                `y` should be column names / keys.

            source_dataframe (None, Pandas DataFrame or dict): Input
                data as Pandas DataFrame or dictionary. If it is a Pandas
                DataFrame, and `x` is None, the index of the dataframe is used
                as `x`

            tick_label (str or array like of str): Labels for each bar. If
                `tick_label` is an array like, its length must be the same as
                the number of bars.
                Consider the argument `label_orientation`

            label_orientation ({'horizontal', 'vertical'}): Orientation of the
                labels

            width (int): The width of the graph, including any axes, titles,
                etc

            height (int): The height of the graph, including any axes, titles,
                etc

            description (str): HTML-formatted text that will be kept bellow the
                graph. It generally includes metadata, details about the,
                graphs, etc. It will be kept when the graph is displayed,
                exported or rendered in a grid with other graphs. If The graph
                is summed with other graph, their metadata will be concatenated
                (adding a new line between both)

            title (str): Title of the graph. The attributes of the title
                (font, etc), cannot be tuned directly in depict

            x_label (None, str): Label of the x-axis

            y_label (None, str): Label of the y-axis

            show_plot (bool): Whether or not the graph must be displayed
                immediately after its creation.
                Notes: The plot object will be returned regardless this
                parameter. Saving the plot is independant from this parameter.

            color (None, str, array-like): If None, the first color of the
                palette is used.
                If string, a color name is expected (hexadecimal, RGB, and
                usual colors are accepted (like 'red', etc)), and it will be
                the same for all the bars.
                If array-like, the length of color should correspond to the
                number of bars, they will correspond to each bar respectively

            colorbar_type ({`auto`, `categorical`, `continuous`}): If 'auto',
                the best type will be chosen wrt the data.
                If `categorical`: a legend will be used, not a colorbar.
                If `continuous`, a colorbar will be displayed on the right
                side of the plot. The data defining the color must be passed
                in `color`

            legend ('auto', array-like): If `auto`, the legend will be set when
                a Pandas DataFrame or a dictionary is provided (the name of the
                columns / keys will be the legend). If array-like, the length
                of `legend` must match with the number of bars drawn

            bar_width ('auto', Number, t-delta, array of number or t-delta):
                The width of the bars (the total width).
                If 'auto': the bar_width will be defined smartly considering
                the minimum distance between 2 bars so that they do not
                overlap.
                If Number: The absolute value that holds for all the bars
                If timedelta: The absolute value that holds for all the bars,
                only for datetime axis.
                If array-like of Number or timedelta: The width of each bar
                is defined separately

            alpha (Number, array-like): Alpha value of the bars.
                If Number: it will the same for all the bars.
                If array-like: its length must match with the number of bars
                drawn. They will correspond to each bar respectively. alpha=1
                means no transparency, alpha=0 means completely transparent.

            x_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly.

            y_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly.

            x_range (array-like): Range of the x-axis

            y_range (array-like): Range of the y-axis

            save_path (None, str): If None, the graph is not saved. If str,
                the graph is saved in html at the given path. If the html
                extension is missing, it will be added automatically

            grid_visible (bool): Whether the background grid must be displayed.
                In depict, you cannot have further control about the background
                grid

        Returns:
            depict.plot
        """
        plot = histogram_base(x=x, y=y, source_dataframe=source_dataframe,
                              tick_label=tick_label,
                              label_orientation=label_orientation, width=width,
                              height=height, description=description,
                              title=title, x_label=x_label, y_label=y_label,
                              show_plot=show_plot, color=color,
                              colorbar_type=colorbar_type, legend=legend,
                              bar_width=bar_width, alpha=alpha,
                              x_axis_type=x_axis_type, y_axis_type=y_axis_type,
                              x_range=x_range, y_range=y_range,
                              grid_visible=grid_visible, session=session,
                              save_path=save_path)
        return plot
    return histogram_updated



from bokeh.io import output_notebook, reset_output


class Session:
    def __init__(self, width, height, save_path, file_exists_mode,
                 description, title, jupyter_notebook,
                 background_color, palette_name, grid_visible, show_plot,
                 width_total_as_session, automatic_color_mapping):
        self.width = width
        self.height = height
        self.save_path = save_path
        self.file_exists_mode = file_exists_mode
        self.description = description
        self.title = title
        if jupyter_notebook:
            output_notebook()
        else:
            reset_output()
        self.background_color = background_color
        self.palette_name = palette_name
        self.grid_visible = grid_visible
        self.show_plot = show_plot
        self.width_total_as_session = width_total_as_session
        self.automatic_color_mapping = automatic_color_mapping



from .plot import Plot
from .tools import show_base, save_base, is_color, format_color, is_iterable
from ..tools.color_palettes import palette_from_name_to_function

import numbers

from bokeh.models import Range1d
from bokeh.plotting import figure
from bokeh.models import ColorBar, LinearColorMapper
import numpy as np
import pandas as pd


def point_base(x, y, source_dataframe, width, height, description, title,
               x_label, y_label, show_plot, color, colorbar_type, legend, size,
               alpha, x_axis_type, y_axis_type, x_range, y_range, grid_visible,
               session, save_path):
    """ Scatter plot

    Args:
        x (array-like): X-axis data
        y (array-like): y-axis data
    """
    # We convert (source, x and y) into only x and y. x and y will be processed
    # normally. source will not be used any more.
    if source_dataframe is not None:
        if isinstance(source_dataframe, dict):
            # TODO: If the user passes a tuple that is not a key, add a
            # message suggesting that he may want to use a list instead
            # because a tuple is considered as a key, and not as an iterable
            # of keys
            # Same for DataFrame
            if y not in source_dataframe.keys():
                raise ValueError('The value of y is not a key of the '
                                 'dictionary provided in source_dataframe')
            if isinstance(legend, str) and (legend.lower() == 'auto'):
                if isinstance(y, str):
                    legend = str(y)
            y = source_dataframe[y]

            if x not in source_dataframe.keys():
                raise ValueError('The value of x is not a key of the '
                                 'dictionary provided in source_dataframe')
            x = source_dataframe[x]

        elif isinstance(source_dataframe, pd.DataFrame):
            if y not in source_dataframe.keys():
                raise ValueError('The value of y is not a key of the '
                                 'DataFrame provided in source_dataframe')
            if isinstance(legend, str) and (legend.lower() == 'auto'):
                if isinstance(y, str):
                    legend = str(y)
            y = source_dataframe[y]

            if x is None:
                # Be careful, the behavior is different for a dict and for a
                # Dataframe
                x = source_dataframe.index.values
            else:
                if x not in source_dataframe.keys():
                    raise ValueError('The value of x is not a key of the '
                                     'DataFrame provided in source_dataframe')
                x = source_dataframe[x].values

    if (x is None) or (y is None):
        raise ValueError('X and y must be specified. They must be a one '
                         'dimensional array like')
    if (not isinstance(x, (list, np.ndarray, tuple)))\
            or (not isinstance(y, (list, np.ndarray, tuple))):
        raise ValueError('X and y must be a one dimensional array like')
    if (np.ndim(x) != 1) or (np.ndim(y) != 1):
        raise ValueError('X and y must be a one dimensional array like')

    # We pre-process `color`
    _color_bar_made = False
    if color is None:
        nb_color_needed = 1
        color = palette_from_name_to_function[session.palette_name](
            nb_color_needed)
        color = [color[0] for _ in y]
    # We pre-process `color`
    # Corner case: if color = [0.5, 0.6, 0.7] and len(y) == 3, we cannot say if
    # color means actually the color defined by [0.5, 0.6, 0.7] or if this
    # quantity should be converted into categorical colors (3 colors here cause
    # the 3 values are different). For that case we use the argument
    # `automatic_color_mapping` in from the session.

    elif (len(color) == 3) \
            and isinstance(color[0], numbers.Real)\
            and session.automatic_color_mapping\
            and (len(y) == 3):
        color = [color, color, color]
    else:
        # General case
        if is_color(color):
            color = [color for _ in y]
        elif isinstance(color, (list, np.ndarray, tuple))\
                and is_color(color[0]):
            if len(color) == len(y):
                pass  # Good case
            elif len(color) == 2:
                # TODO: Implement a linear set of color between the 2 given
                #  colors
                raise ValueError('Number of elements in `color` invalid')
            else:
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')
        elif isinstance(color, (list, np.ndarray, tuple)) and isinstance(
                color[0], numbers.Real):
            if len(color) != len(y):
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')

            if not colorbar_type:
                colorbar_type = 'auto'
                _add_color_bar = False
            else:
                _add_color_bar = True

            if colorbar_type.lower() == 'auto':
                if ((len(np.unique(color)) / len(color)) <= 0.5)\
                        and (len(np.unique(color) < 9)):
                    # Data are considered categorical
                    colorbar_type = 'categorical'
                else:
                    colorbar_type = 'continuous'
            else:
                colorbar_type = colorbar_type.lower()

            if colorbar_type == 'categorical':
                if isinstance(legend, str) and (legend.lower() == 'auto'):
                    legend = [str(c) for c in color]
                color_unique = sorted(list(np.unique(color)))
                nb_color_needed = len(color_unique)
                palette_colors = palette_from_name_to_function[
                    session.palette_name](nb_color_needed)
                color = [palette_colors[color_unique.index(c)] for c in color]
            elif colorbar_type == 'continuous':
                palette_colors = palette_from_name_to_function[
                    session.palette_name](256)
                color_mapper = LinearColorMapper(palette=palette_colors,
                                                 low=np.min(color),
                                                 high=np.max(color))
                color = np.array(color)
                col_indexes = (color - color.min()) / (
                            (color.max() - color.min()) / (256 - 1))
                color_min = np.min(color)
                color_max = np.max(color)
                color = [palette_colors[int(ci)] for ci in col_indexes]
                if _add_color_bar:
                    color_bar = ColorBar(color_mapper=color_mapper,
                                         location=(0, 0))
                    _color_bar_made = True

    color = [format_color(c) for c in color]

    # We pre-process `legend`
    # add_legend = True
    if (legend is None) or (legend == 'auto'):
        legend = ['' for _ in y]
        # add_legend = False
    elif isinstance(legend, str):
        legend = [legend for _ in y]
    elif isinstance(legend, (list, np.ndarray, tuple)):
        if len(legend) != len(y):
            raise ValueError('The number of elements in `legend` is not '
                             'consistent with the data')

    # We pre-process `size`
    if isinstance(size, numbers.Real):
        size = [size for _ in y]
    elif isinstance(size, (list, np.ndarray, tuple)):
        if len(size) != len(y):
            raise ValueError('The size argument given is non consistent '
                             'with the data')

    # We pre-process `alpha`
    if isinstance(alpha, numbers.Real):
        alpha = [alpha for _ in y]
    elif isinstance(alpha, (list, np.ndarray, tuple)):
        if len(alpha) != len(y):
            raise ValueError(
                'The alpha argument given is non consistent with the data')

    # We pre-process `x_axis_type` and `y_axis_type`
    if x_axis_type.lower() == 'auto':
        if isinstance(x[0], numbers.Real):
            x_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(x[0], errors='raise')
            except ValueError:
                raise ValueError('`x_axis_type` is set to `auto` and x[0] '
                                 'is neither a number, nor an object parsable '
                                 'as a date. Object x[0]: {}'.format(x[0]))
            x_axis_type = 'datetime'
            x = pd.to_datetime(x)
    elif x_axis_type.lower() in ['numeric', 'numerical']:
        x_axis_type = 'linear'
    elif x_axis_type.lower() in ['date', 'datetime', 'time']:
        x_axis_type = 'datetime'
        x = pd.to_datetime(x)
    # Same with y axis
    if y_axis_type.lower() == 'auto':
        if isinstance(y[0], numbers.Real):
            y_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(y[0], errors='raise')
            except ValueError:
                raise ValueError('`y_axis_type` is set to `auto` and y[0] '
                                 'is neither a number, nor an object parsable '
                                 'as a date. Object y[0]: {}'.format(y[0]))
            y_axis_type = 'datetime'
            y = pd.to_datetime(y)
    elif y_axis_type.lower() in ['numeric', 'numerical']:
        y_axis_type = 'linear'
    elif y_axis_type.lower() in ['date', 'datetime', 'time']:
        y_axis_type = 'datetime'
        y = pd.to_datetime(y)

    # We group x and y based on legend because in bokek, figure.scatter can
    # only set one legend label by scatter plot. So if the legend contains
    # an iterable of strings, we need to look for unique values and mask
    # all attributes
    legend_unique = np.unique(legend)
    mask_all = [l_i == np.array(legend) for l_i in legend_unique]
    x = [np.array(x)[mask] for mask in mask_all]
    y = [np.array(y)[mask] for mask in mask_all]
    color = [np.array(color)[mask] for mask in mask_all]
    if isinstance(color[0][0], np.ndarray):
        color = [[tuple(c_i_i) for c_i_i in c_i] for c_i in color]
    size = [np.array(size)[mask] for mask in mask_all]
    alpha = [np.array(alpha)[mask] for mask in mask_all]

    steps = []
    legend_exist = False
    legend_unique = [str(lu) for lu in legend_unique]
    for (x_i, y_i, col_i, leg_i, s_i, a_i) in zip(x, y, color, legend_unique,
                                                  size, alpha):
        if leg_i:
            legend_exist = True

            def step(f, x_copy=x_i, y_copy=y_i, col_c=col_i, leg_c=leg_i,
                     s_c=s_i, a_c=a_i):
                f.scatter(x=x_copy, y=y_copy, color=col_c, legend_label=leg_c,
                          size=s_c, alpha=a_c)
        else:
            def step(f, x_copy=x_i, y_copy=y_i, col_c=col_i,
                     s_c=s_i, a_c=a_i):
                f.scatter(x=x_copy, y=y_copy, color=col_c, size=s_c, alpha=a_c)
        steps.append(step)

    if legend_exist:
        def make_legend_interactive(f):
            f.legend.click_policy = "hide"
        steps.append(make_legend_interactive)

    def _make_fig():
        fig = figure(width=width, height=height, title=title,
                     background_fill_color=session.background_color,
                     x_axis_label=x_label, y_axis_label=y_label,
                     x_axis_type=x_axis_type, y_axis_type=y_axis_type)
        fig.title.align = 'center'
        fig.title.text_color = '#33331a'
        fig.xgrid.grid_line_dash = [8, 3, 2, 3]
        fig.ygrid.grid_line_dash = [8, 3, 2, 3]
        fig.toolbar.autohide = True
        if x_range is not None:
            if not is_iterable(x_range):
                raise ValueError('`x_range` argument should be an iterable')
            fig.x_range = Range1d(x_range[0], x_range[1])
        if y_range is not None:
            if not is_iterable(y_range):
                raise ValueError('`y_range` argument should be an iterable')
            fig.y_range = Range1d(y_range[0], y_range[1])
        if _color_bar_made and _add_color_bar:
            color_mapper = LinearColorMapper(palette=palette_colors,
                                             low=color_min,
                                             high=color_max)
            c_bar = ColorBar(color_mapper=color_mapper, location=(0, 0))
            fig.add_layout(c_bar, 'right')
        return fig

    plot = Plot(make_figure=_make_fig, steps=steps, description=description,
                figure=_make_fig(), width=width, grid_visible=grid_visible,
                width_session=session.width, session=session)

    if save_path:
        save_base(plot=plot, save_path=save_path,
                  file_exists_mode=session.file_exists_mode,
                  width_total_as_session=False, share_x=False, share_y=False)

    if show_plot:
        return show_base(plot) or plot
    else:
        return plot


def _update_point_default_args(point, session):
    def point_updated(x, y, source_dataframe=None, width=session.width,
                      height=session.height, description=session.description,
                      title=session.title, x_label=None, y_label=None,
                      show_plot=session.show_plot, color=None,
                      colorbar_type='auto', legend='auto', size=6, alpha=1,
                      x_axis_type='auto', y_axis_type='auto', x_range=None,
                      y_range=None, save_path=session.save_path,
                      grid_visible=session.grid_visible):
        """Plot a graph with points (scatter-plot)

        Args:
            x (array-like of dimension 1): The x-coordinates for the points.
                The values can be either numbers, or dates (datetimes or
                parsable strings). If a `source_dataframe` is given, `x` and
                `y` should be column names / keys. `x` can be None is a special
                case: when a pandas dataframe is given as `source_dataframe`
                and you want to use the index of the dataframe as `x`

            y (array-like of dimension 1): The y-coordinates for the points.
                The values can be either numbers, or dates (datetimes or
                parsable strings). If a `source_dataframe` is given, `x` and
                `y` should be column names / keys.

            source_dataframe (None, Pandas DataFrame or dict): Input
                data as Pandas DataFrame or dictionary. If it is a Pandas
                DataFrame, and `x` is None, the index of the dataframe is used
                as `x`

            width (int): The width of the graph, including any axes, titles,
                etc

            height (int): The height of the graph, including any axes, titles,
                etc

            description (str): HTML-formatted text that will be kept bellow the
                graph. It generally includes metadata, details about the,
                graphs, etc. It will be kept when the graph is displayed,
                exported or rendered in a grid with other graphs. If The graph
                is summed with other graph, their metadata will be concatenated
                (adding a new line between both)

            title (str): Title of the graph. The attributes of the title
                (font, etc), cannot be tuned directly in depict

            x_label (None, str): Label of the x-axis

            y_label (None, str): Label of the y-axis

            show_plot (bool): Whether or not the graph must be displayed
                immediatly after its creation

            color (None, str, array-like): If None, the first color of the
                palette is used.
                If string, a color name is expected (hexadecimal, RGB, and
                usual colors are accepted (like 'red', etc)), and it will be
                the same for all the points.
                If array-like, the length of color should correspond to the
                number of points drawn, they will correspond to each point
                respectively

            colorbar_type ({`auto`, `categorical`, `continuous`}): If 'auto',
                the best type will be chosen wrt the data.
                If `categorical`: a legend will be used, not a colorbar.
                If `continuous`, a colorbar will be displayed on the right
                side of the plot. The data defining the color must be passed
                in `color`

            legend ('auto', array-like): If `auto`, the legend will be set when
                a Pandas DataFrame or a dictionary is provided (the name of the
                columns / keys will be the legend). If array-like, the length
                of `legend` must match with the number of points drawn

            size (Number, array-like): Size of the points. If Number
                it will the same for all the points. If it is an array-like,
                its length must match with the number of points drawn. They
                will correspond to each point respectively. A size of 6 is a
                reasonable default value

            alpha (Number, array-like): Alpha value of the points. If Number
                it will the same for all the points. If it is an array-like,
                its length must match with the number of points drawn. They
                will correspond to each line respectively. alpha=1 means no
                transparency, alpha=0 means completely transparent.

            x_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly.

            y_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly.

            x_range (array-like): Range of the x-axis

            y_range (array-like): Range of the y-axis

            save_path (None, str): If None, the graph is not saved. If str,
                the graph is saved in html at the given path. If the html
                extension is missing, it will be added automatically

            grid_visible (bool): Whether the background grid must be displayed.
                In depict, you cannot have further control about the background
                grid

        Returns:
            depict.plot
        """
        plot = point(x=x, y=y, source_dataframe=source_dataframe, width=width,
                     height=height, description=description, title=title,
                     x_label=x_label, y_label=y_label, show_plot=show_plot,
                     color=color, colorbar_type=colorbar_type, legend=legend,
                     size=size, alpha=alpha, x_axis_type=x_axis_type,
                     y_axis_type=y_axis_type, x_range=x_range, y_range=y_range,
                     grid_visible=grid_visible, session=session,
                     save_path=save_path)
        return plot
    return point_updated



class Plot:
    def __init__(self, make_figure, steps, description, figure, width,
                 grid_visible, width_session, session):
        self.make_figure = make_figure
        self.steps = steps
        self.description = description
        self.figure = figure
        self.width = width
        self.grid_visible = grid_visible
        self.width_session = width_session
        self.session = session

    def __add__(self, other):
        new_description = self.description + '<br>' + other.description
        fig_sum = self.make_figure()
        for step in self.steps + other.steps:
            step(fig_sum)
        return Plot(make_figure=self.make_figure,
                    steps=self.steps + other.steps,
                    description=new_description,
                    figure=fig_sum, width=self.width,
                    grid_visible=self.grid_visible,
                    width_session=self.width_session,
                    session=self.session)



from .core.histogram import histogram_base as _histogram_base
from .core.histogram import _update_histogram_default_args
from .core.line import line_base as _line_base
from .core.line import _update_line_default_args

from .core.point import point_base as _point_base
from .core.point import _update_point_default_args

from .core.tools import save_base as _save_base
from .core.tools import _update_save_default_args

from .core.tools import show_base as _show_base
from .core.tools import _update_show_default_args

from .core.session import Session as _Session


__all__ = ['session', 'line', 'point', 'histogram', 'show', 'save']


def session(width=900, height=400, save_path=None, file_exists_mode='append',
            description='', title='', jupyter_notebook=False,
            background_color='aliceblue', palette_name='categories_10',
            grid_visible=True, show_plot=True, width_total_as_session=True,
            automatic_color_mapping=True):
    global _SESSION, histogram, line, point, save, show
    _SESSION = _Session(width=width, height=height, save_path=save_path,
                        file_exists_mode=file_exists_mode,
                        description=description, title=title,
                        jupyter_notebook=jupyter_notebook,
                        background_color=background_color,
                        palette_name=palette_name,
                        grid_visible=grid_visible, show_plot=show_plot,
                        width_total_as_session=width_total_as_session,
                        automatic_color_mapping=automatic_color_mapping)

    save = _update_save_default_args(save_base=_save_base, session=_SESSION)
    histogram = _update_histogram_default_args(histogram_base=_histogram_base,
                                               session=_SESSION)
    line = _update_line_default_args(line=_line_base, session=_SESSION)
    point = _update_point_default_args(point=_point_base, session=_SESSION)
    show = _update_show_default_args(show_base=_show_base, session=_SESSION)


session()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="depict",
    version="1.0.3",
    author="Victor Boulanger",
    author_email="vb@live.fr",
    description="Business grade visualizations in seconds",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vboulanger/depict",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    python_requires='>=3.5',
    install_requires=['bokeh>=2.0.0', 'seaborn>=0.9.0', 'pandas>=0.24.2'],
    extras_require={'test': ['pytest', 'pycodestyle'],
                    'doc': ['Sphinx']},
    tests_require=['pytest'],
    setup_requires=['pytest-runner'],
)

# https://pypi.org/classifiers/



from .plot import Plot
from .tools import show_base, save_base, is_color, format_color, is_iterable
from ..tools.color_palettes import palette_from_name_to_function

import numbers

from bokeh.models import Range1d
from bokeh.plotting import figure
from bokeh.models import ColorBar, LinearColorMapper
import numpy as np
import pandas as pd


def line_base(y, x, source_dataframe, width, height, description, title,
              x_label, y_label, show_plot, color, colorbar_type, legend,
              line_width, alpha, style, x_axis_type, y_axis_type, x_range,
              y_range, fill_between, grid_visible, session, save_path):
    """ One dimensional plot. This is the docstring of line

    Args:
        x (array-like): X-axis data
        y (array-like): y-axis data
        x_range (array-like) Range of the x-axis. E.g. `[0, 1]`
        y_range (array-like) Range of the y-axis. E.g. `[0, 1]`
    """
    # We convert (source, x and y) into only x and y. x and y will be processed
    # normally. source will not be used any more.
    if source_dataframe is not None:
        if isinstance(source_dataframe, dict):
            # TODO: If the user passes a tuple that is not a key, add a
            # message suggesting that he may want to use a list instead
            # because a tuple is considered as a key, and not as an iterable
            # of keys
            # Same for DataFrame

            if isinstance(y, (list, np.ndarray)):
                if not all([y_i in source_dataframe.keys() for y_i in y]):
                    raise ValueError('At least some of the values of y are '
                                     'not keys of the dictionary provided in '
                                     'source_dataframe')
                if legend.lower() == 'auto':
                    legend = [y_i if isinstance(y_i, str) else '' for y_i in y]
                y = [source_dataframe[y_i] for y_i in y]
            else:
                if y not in source_dataframe.keys():
                    raise ValueError('The value of y is not a key of the '
                                     'dictionary provided in source_dataframe')
                if legend.lower() == 'auto':
                    if isinstance(y, str):
                        legend = str(y)
                y = [source_dataframe[y]]

            if x is None:
                pass  # Will be treated as the default x behavior
            elif isinstance(x, (list, np.ndarray)):
                if not all([x_i in source_dataframe.keys() for x_i in x]):
                    raise ValueError('At least some of the values of x are '
                                     'not keys of the dictionary provided in '
                                     'source_dataframe')
                x = [source_dataframe[x_i] for x_i in x]
            else:
                if x not in source_dataframe.keys():
                    raise ValueError('The value of x is not a key of the '
                                     'dictionary provided in source_dataframe')
                x = source_dataframe[x]

        elif isinstance(source_dataframe, pd.DataFrame):
            if isinstance(y, (list, np.ndarray)):
                if not all([y_i in source_dataframe.keys() for y_i in y]):
                    raise ValueError('At least some of the values of y are '
                                     'not keys of the DataFrame provided in '
                                     'source_dataframe')
                if legend.lower() == 'auto':
                    legend = [y_i if isinstance(y_i, str) else '' for y_i in y]
                y = [source_dataframe[y_i].values for y_i in y]
            else:
                if y not in source_dataframe.keys():
                    raise ValueError('The value of y is not a key of the '
                                     'DataFrame provided in source_dataframe')
                if legend.lower() == 'auto':
                    if isinstance(y, str):
                        legend = str(y)
                y = [source_dataframe[y]]

            if x is None:
                # Be careful, the behavior is different for a dict and for a
                # Dataframe
                x = source_dataframe.index.values
            elif isinstance(x, (list, np.ndarray)):
                if not all([x_i in source_dataframe.keys() for x_i in x]):
                    raise ValueError('At least some of the values of x are '
                                     'not keys of the DataFrame provided in '
                                     'source_dataframe')
                x = [source_dataframe[x_i].values for x_i in x]
            else:
                if x not in source_dataframe.keys():
                    raise ValueError('The value of x is not a key of the '
                                     'DataFrame provided in source_dataframe')
                x = source_dataframe[x].values

    # We pre-process `x` and `y`
    if x is None:
        if isinstance(y[0], (list, np.ndarray, tuple)):
            x = [np.arange(len(y_i)) for y_i in y]
        else:
            x = [np.arange(len(y))]
            y = [y]
    elif np.ndim(x) == 1:
        if np.ndim(y) == 1:
            x = [x]
            y = [y]
        elif np.ndim(y) == 2:
            x = [x for _ in y]
    elif np.ndim(x) == 2:
        if np.ndim(y) == 1:  # Corner case
            y = [y for _ in x]

    # We pre-process `color`
    _color_bar_made = False
    if color is None:
        nb_color_needed = len(y)
        color = palette_from_name_to_function[session.palette_name](
            nb_color_needed)
    # We pre-process `color`
    # Corner case: if color = [0.5, 0.6, 0.7] and len(y) == 3, we cannot say if
    # color means actually the color defined by [0.5, 0.6, 0.7] or if this
    # quantity should be converted into categorical colors (3 colors here cause
    # the 3 values are different). For that case we use the argument
    # `automatic_color_mapping` in from the session.

    elif (len(color) == 3) \
            and isinstance(color[0], numbers.Real)\
            and session.automatic_color_mapping\
            and (len(y) == 3):
        color = [color, color, color]
    else:
        # General case
        if is_color(color):
            color = [color for _ in y]
        elif isinstance(color, (list, np.ndarray, tuple))\
                and is_color(color[0]):
            if len(color) == len(y):
                pass  # Good case
            elif len(color) == 2:
                # TODO: Implement a linear set of color between the 2 given
                #  colors
                raise ValueError('Number of elements in `color` invalid')
            else:
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')
        elif isinstance(color, (list, np.ndarray, tuple)) and isinstance(
                color[0], numbers.Real):
            if len(color) != len(y):
                # TODO: Improve error message
                raise ValueError('Number of elements in `color` invalid')

            if not colorbar_type:
                colorbar_type = 'auto'
                _add_color_bar = False
            else:
                _add_color_bar = True

            if colorbar_type.lower() == 'auto':
                if ((len(np.unique(color)) / len(color)) <= 0.5) and (
                  len(np.unique(color) < 9)):
                    # Data are considered categorical
                    colorbar_type = 'categorical'
                else:
                    colorbar_type = 'continuous'
            else:
                colorbar_type = colorbar_type.lower()

            if colorbar_type == 'categorical':
                if isinstance(legend, str) and (legend.lower() == 'auto'):
                    legend = [str(c) for c in color]
                color_unique = sorted(list(np.unique(color)))
                nb_color_needed = len(color_unique)
                palette_colors = palette_from_name_to_function[
                    session.palette_name](nb_color_needed)
                color = [palette_colors[color_unique.index(c)] for c in color]
            elif colorbar_type == 'continuous':
                palette_colors = palette_from_name_to_function[
                    session.palette_name](256)
                color_mapper = LinearColorMapper(palette=palette_colors,
                                                 low=np.min(color),
                                                 high=np.max(color))
                color = np.array(color)
                col_indexes = (color - color.min()) / (
                            (color.max() - color.min()) / (256 - 1))
                color_min = np.min(color)
                color_max = np.max(color)
                color = [palette_colors[int(ci)] for ci in col_indexes]
                if _add_color_bar:
                    color_bar = ColorBar(color_mapper=color_mapper,
                                         location=(0, 0))
                    _color_bar_made = True

    color = [format_color(c) for c in color]

    # We pre-process `legend`
    # add_legend = True
    if (legend is None) or (legend == 'auto'):
        legend = ['' for _ in y]
        # add_legend = False
    elif isinstance(legend, str):
        legend = [legend for _ in y]
    elif isinstance(legend, (list, np.ndarray, tuple)):
        if len(legend) != len(y):
            raise ValueError('The number of elements in `legend` is not '
                             'consistent with the data')

    # We pre-process `line_width`
    if isinstance(line_width, numbers.Real):
        line_width = [line_width for _ in y]
    elif isinstance(line_width, (list, np.ndarray, tuple)):
        if len(line_width) != len(y):
            raise ValueError('The line_width argument given is non consistent '
                             'with the data')

    # We pre-process `alpha`
    if isinstance(alpha, numbers.Real):
        alpha = [alpha for _ in y]
    elif isinstance(alpha, (list, np.ndarray, tuple)):
        if len(alpha) != len(y):
            raise ValueError(
                'The alpha argument given is non consistent '
                'with the data')

    # We pre-process `style`
    if isinstance(style, str):
        style = [style.lower() for _ in y]
    elif isinstance(style, (list, np.ndarray, tuple)):
        if len(style) != len(y):
            raise ValueError(
                'The style argument given is non consistent '
                'with the data')
        style = [s.lower() for s in style]

    def guess_style(style):
        # This list is to be completted to match as much as possible with
        # Matplotlib equivalent attribute
        if style in ['-', '--', '---']:
            return 'dashed'
        elif style in ['.', '..', '...']:
            return 'dotted'
        elif style == '.-':
            return 'dotdash'
        elif style == '-.':
            return 'dashdot'
        else:
            return style
    style = [guess_style(s) for s in style]

    # We pre-process `x_axis_type` and `y_axis_type`
    if x_axis_type.lower() == 'auto':
        if isinstance(x[0][0], numbers.Real):
            x_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(x[0][0], errors='raise')
            except ValueError:
                # TODO: Change message error can x[0][0] may not make sense for
                # the user. Same for y axis below
                error_msg = (
                    '`x_axis_type` is set to `auto` and x[0][0] '
                    'is neither a number, nor an object parsable '
                    'as a date. Object x[0][0]: {}'
                ).format(x[0][0])
                raise ValueError(error_msg)
            x_axis_type = 'datetime'
            x = [pd.to_datetime(x_i) for x_i in x]
    elif x_axis_type.lower() in ['numeric', 'numerical']:
        x_axis_type = 'linear'
    elif x_axis_type.lower() in ['date', 'datetime', 'time']:
        x_axis_type = 'datetime'
        x = [pd.to_datetime(x_i) for x_i in x]
    # Same with y axis
    if y_axis_type.lower() == 'auto':
        if isinstance(y[0][0], numbers.Real):
            y_axis_type = 'linear'
        else:
            try:
                pd.to_datetime(y[0][0], errors='raise')
            except ValueError:
                error_msg = (
                    '`y_axis_type` is set to `auto` and y[0][0] '
                    'is neither a number, nor an object parsable '
                    'as a date. Object y[0][0]: {}'
                ).format(y[0][0])
                raise ValueError(error_msg)
            y_axis_type = 'datetime'
            y = [pd.to_datetime(y_i) for y_i in y]
    elif y_axis_type.lower() in ['numeric', 'numerical']:
        y_axis_type = 'linear'
    elif y_axis_type.lower() in ['date', 'datetime', 'time']:
        y_axis_type = 'datetime'
        y = [pd.to_datetime(y_i) for y_i in y]

    steps = []
    legend_exist = False
    for (x_i, y_i, col_i, leg_i, lw_i, a_i, s_i) in zip(x, y, color, legend,
                                                        line_width, alpha,
                                                        style):
        if leg_i:
            legend_exist = True

            def step(f, x_copy=x_i, y_copy=y_i, col_c=col_i, leg_c=leg_i,
                     lw_c=lw_i, a_c=a_i, s_c=s_i):
                f.line(x=x_copy, y=y_copy, color=col_c, legend_label=leg_c,
                       line_width=lw_c, alpha=a_c, line_dash=s_c)
        else:
            def step(f, x_copy=x_i, y_copy=y_i, col_c=col_i, lw_c=lw_i,
                     a_c=a_i, s_c=s_i):
                f.line(x=x_copy, y=y_copy, color=col_c, line_width=lw_c,
                       alpha=a_c, line_dash=s_c)
        steps.append(step)

    if legend_exist:
        def make_legend_interactive(f):
            f.legend.click_policy = "hide"
        steps.append(make_legend_interactive)

    if fill_between:
        if len(y) == 1:
            pass
            # TODO: Add warning and a piece of code to explain how to fill the
            #  area between 2 curves
        elif len(y) >= 2:
            if not all(x[0] == x[1]):
                # TODO: Transform this error into a warning and explain better
                raise ValueError('When `fill_between` is True, x[0] and x[1] '
                                 'must be the same.')

            def fill_between_curves(f, x_0=x[0], y_0=y[0], y_1=y[1]):
                f.varea(x=x_0, y1=y_0, y2=y_1, alpha=alpha[0], color=color[0])
            steps.append(fill_between_curves)

    def _make_fig():
        fig = figure(width=width, height=height, title=title,
                     background_fill_color=session.background_color,
                     x_axis_label=x_label, y_axis_label=y_label,
                     x_axis_type=x_axis_type, y_axis_type=y_axis_type)
        fig.title.align = 'center'
        fig.title.text_color = '#33331a'
        fig.xgrid.grid_line_dash = [8, 3, 2, 3]
        fig.ygrid.grid_line_dash = [8, 3, 2, 3]
        fig.toolbar.autohide = True
        if x_range is not None:
            if not is_iterable(x_range):
                raise ValueError('`x_range` argument should be an iterable')
            fig.x_range = Range1d(x_range[0], x_range[1])
        if y_range is not None:
            if not is_iterable(y_range):
                raise ValueError('`y_range` argument should be an iterable')
            fig.y_range = Range1d(y_range[0], y_range[1])
        if _color_bar_made and _add_color_bar:
            color_mapper = LinearColorMapper(palette=palette_colors,
                                             low=color_min,
                                             high=color_max)
            c_bar = ColorBar(color_mapper=color_mapper, location=(0, 0))
            fig.add_layout(c_bar, 'right')
        return fig

    plot = Plot(make_figure=_make_fig, steps=steps, description=description,
                figure=_make_fig(), width=width, grid_visible=grid_visible,
                width_session=session.width, session=session)

    if save_path:
        save_base(plot=plot, save_path=save_path,
                  file_exists_mode=session.file_exists_mode,
                  width_total_as_session=False, share_x=False, share_y=False)

    if show_plot:
        return show_base(plot) or plot
    else:
        return plot


def _update_line_default_args(line, session):
    def line_updated(y, x=None, source_dataframe=None, width=session.width,
                     height=session.height, description=session.description,
                     title=session.title, x_label=None, y_label=None,
                     show_plot=session.show_plot, color=None,
                     colorbar_type='auto', legend='auto', line_width=1,
                     alpha=1, style='solid', x_axis_type='auto',
                     y_axis_type='auto', x_range=None, y_range=None,
                     fill_between=False, save_path=session.save_path,
                     grid_visible=session.grid_visible):
        """Plot a graph with one-dimensional line(s)

        Args:
            y (array-like of dimension 1 or 2): The y-coordinates for the
                points of the line(s). If `y` is 1-d, one line will be drawn,
                if  it is 2-d, a set of lines will be drawn.
                The values can be either numbers, or dates (datetimes or
                parsable strings). If a `source_dataframe` is given, `x` and
                `y` should be column names / keys.

            x (None, array-like of dimension 1 or 2): The x-coordinates for the
                points of the line(s). If `None`, indexes starting from 0 and
                adapted to `y` will be used. If `y` is 1-d, one line will be
                drawn, if  it is 2-d, a set of lines will be drawn.
                The values can be either numbers, or dates (datetimes or
                parsable strings). If a `source_dataframe` is given, `x` and
                `y` should be column names / keys.

            source_dataframe (None, Pandas DataFrame or dict): Input data as
                Pandas DataFrame or dictionary. If it is a Pandas DataFrame,
                and `x` is None, the index of the dataframe is used as `x`

            width (int): The width of the graph, including any axes, titles,
                etc

            height (int): The height of the graph, including any axes, titles,
                etc

            description (str): HTML-formatted text that will be kept bellow the
                graph. It generally includes metadata, details about the,
                graphs, etc. It will be kept when the graph is displayed,
                exported or rendered in a grid with other graphs. If The graph
                is summed with other graph, their metadata will be concatenated
                (adding a new line between both)

            title (str): Title of the graph. The attributes of the title
                (font), cannot be tuned directly in depict

            x_label (None, str): Label of the x-axis

            y_label (None, str): Label of the y-axis

            show_plot (bool): Whether or not the graph must be displayed
                immediatly after its creation

            color (None, str, array-like): If None, the first color of the
                palette is used.
                If string, a color name is expected (hexadecimal, RGB, and
                usual colors are accepted), and it will be the same for all
                the lines.
                If array-like, the length of color should correspond to the
                number of lines drawn, they will correspond to each line
                respectively

            colorbar_type ({`auto`, `categorical`, `continuous`}): If 'auto',
                the best type will be chosen wrt the data. If `categorical`:
                a legend will be used, not a colorbar. If `continuous`, a
                colorbar will be displayed on the right side of the plot.
                The data defining the color must be passed in `color`

            legend ('auto', array-like): If `auto`, the legend will be set when
                a Pandas DataFrame ora dictionary is provided (the name of the
                columns / keys will be the legend). If array-like, the length
                of `legend` must match with the number of curves dranw

            line_width (Number, array-like): Width of the line(s). If Number
                it will the same for all the lines. If it is an array-like,
                its length must match with the number of lines drawn. They will
                correspond to each line respectively

            alpha (Number, array-like): Alpha value of the line(s). If Number
                it will the same for all the lines. If it is an array-like,
                its length must match with the number of lines drawn. They will
                correspond to each line respectively

            style ({'solid', 'dashed', 'dotted', 'dotdash', dashdot'} or array
                like of those): Style of the line(s). If it is an array-like,
                its length must match with the number of lines drawn. They will
                correspond to each line respectively

            x_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly

            y_axis_type ({'auto', 'numerical', 'datetime'}): Type of the axis.
                If 'auto', the type will be set automatically based on the data
                provided. If 'numerical', 'datetime', the type is set
                accordingly

            x_range (array-like): Range of the x-axis

            y_range (array-like): Range of the y-axis

            fill_between (bool): If True, there must be at least 2 curves to
                plot and the area between the first 2 curves is filled. The
                color of the first line and its alpha value are used for the
                part filled. In depict, you cannot have more control about
                this area

            save_path (None, str): If None, the graph is not saved. If str,
                the graph is saved in html at the gicen path. If the html
                extension is missing, it will be added automatically

            grid_visible (bool): Whether the background grid must be displayed.
                In depict, you cannot have further control about the background
                grid

        Returns:
            None
        """
        plot = line(y=y, x=x, source_dataframe=source_dataframe, width=width,
                    height=height, description=description, title=title,
                    x_label=x_label, y_label=y_label, show_plot=show_plot,
                    color=color, colorbar_type=colorbar_type, legend=legend,
                    line_width=line_width, alpha=alpha, style=style,
                    x_axis_type=x_axis_type, y_axis_type=y_axis_type,
                    x_range=x_range, y_range=y_range,
                    fill_between=fill_between, grid_visible=grid_visible,
                    session=session, save_path=save_path)
        return plot
    return line_updated
